// -*-c++-*-

/*!
  \file body_clear_ball.cpp
  \brief kick the ball to escape from a dangerous situation
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "body_clear_clone.h"

#include <rcsc/action/intention_kick.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_kick_two_step.h>
#include <rcsc/action/body_kick_multi_step.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/line_2d.h>

#include <rcsc/player/say_message_builder.h>


const double Body_ClearClone::S_SEARCH_ANGLE = 8.0;

rcsc::GameTime Body_ClearClone::S_last_calc_time( 0, 0 );
rcsc::AngleDeg Body_ClearClone::S_cached_best_angle = 0.0;

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Body_ClearClone::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                  "%s:%d: Body_ClearClone"
                  ,__FILE__, __LINE__ );

    if ( ! agent->world().self().isKickable() )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " not ball kickable!"
                  << std::endl;
        rcsc::dlog.addText( rcsc::Logger::ACTION,
                      "%s:%d:  not kickable"
                      ,__FILE__, __LINE__ );
        return false;
    }

    if ( S_last_calc_time != agent->world().time() )
    {
        rcsc::dlog.addText( rcsc::Logger::CLEAR,
                      "%s:%d: update clear angle"
                      ,__FILE__, __LINE__ );

        double lower_angle, upper_angle;

        if ( agent->world().self().pos().y > rcsc::ServerParam::i().goalHalfWidth() - 1.0 )
        {
            lower_angle = 0.0;
            upper_angle = 90.0;
        }
        else if ( agent->world().self().pos().y < -rcsc::ServerParam::i().goalHalfWidth() + 1.0 )
        {
            lower_angle = -90.0;
            upper_angle = 0.0;
        }
        else
        {
            lower_angle = -60.0;
            upper_angle = 60.0;
        }

        S_cached_best_angle = get_best_angle( agent,
                                              lower_angle,
                                              upper_angle,
                                              ! agent->world().self().goalie() );
        S_last_calc_time = agent->world().time();
    }


    const rcsc::Vector2D target_point
        = agent->world().self().pos()
        + rcsc::Vector2D::polar2vector(30.0, S_cached_best_angle);

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                  "%s:%d: clear angle = %f"
                  ,__FILE__, __LINE__,
                  S_cached_best_angle.degree() );

    agent->debugClient().setTarget( target_point );

    if ( agent->world().self().goalie()
         || agent->world().gameMode().type() == rcsc::GameMode::GoalKick_ )
    {
        rcsc::Body_KickOneStep( target_point,
                          rcsc::ServerParam::i().ballSpeedMax()
                          ).execute( agent );
        agent->debugClient().addMessage( "Clear" );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                      "%s:%d: goalie or goal_kick register clear kick. one step."
                      ,__FILE__, __LINE__ );

        //calc nearest teammate
        rcsc::Circle2D circle( agent->world().self().pos()
                               + rcsc::Vector2D::polar2vector(15.0,S_cached_best_angle),
                               14.5);
        agent->debugClient().addCircle( circle );
        double min_dist=100,tmp_dist;
        int nearest_unum=-1;
        rcsc::Vector2D nearest_point;
        
        const rcsc::PlayerPtrCont::const_iterator end
            = agent->world().teammatesFromSelf().end();
        for ( rcsc::PlayerPtrCont::const_iterator
                  it = agent->world().teammatesFromSelf().begin();
              it != end;
              ++it )
        {
            //check unum
            if( (*it)->unum() <= 0 || (*it)->unum() > 11 )
                continue;

            //check if circle contains
            if( !circle.contains( (*it)->pos() ) )
                continue;

            //calc dist
            rcsc::Line2D line(agent->world().self().pos(),S_cached_best_angle);
            tmp_dist = line.dist( (*it)->pos() );
            if( min_dist > tmp_dist ){
                min_dist = tmp_dist;
                nearest_unum = (*it)->unum();
                nearest_point = line.projection( (*it)->pos() );
            }
        }

        if( min_dist != 100 ){
            agent->addSayMessage( new rcsc::PassMessage( nearest_unum,
                                                   nearest_point,
                                                   agent->world().ball().pos(),
                                                   agent->world().ball().vel()));
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: nearest unum %d, point(%f,%f)"
                                ,__FILE__, __LINE__,nearest_unum,
                                nearest_point.x,nearest_point.y);
        }
                                                   
        
    }
    else
    {
        rcsc::Vector2D one_step_vel
            = rcsc::Body_KickOneStep::get_max_possible_vel
            ( ( target_point - agent->world().ball().pos() ).th(),
              agent->world().self().kickRate(),
              agent->world().ball().vel() );

        if ( one_step_vel.r() > 2.0 )
        {
            rcsc::Body_KickOneStep( target_point,
                              rcsc::ServerParam::i().ballSpeedMax()
                              ).execute( agent );
            agent->debugClient().addMessage( "Clear1K" );
            return true;
        }

        agent->debugClient().addMessage( "Clear2K" );
        rcsc::Body_KickTwoStep( target_point,
                          rcsc::ServerParam::i().ballSpeedMax(),
                          false //true // enforce
                          ).execute( agent );

        agent->setIntention
            ( new rcsc::IntentionKick( target_point,
                                       rcsc::ServerParam::i().ballSpeedMax(),
                                 1,
                                 true, // enforde
                                 agent->world().time() ) );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                      "%s:%d: register clear kick intention"
                      ,__FILE__, __LINE__ );
    }
    return true;
}

/*-------------------------------------------------------------------*/
/*!
  lower < uppeer
*/
rcsc::AngleDeg
Body_ClearClone::get_best_angle( const rcsc::PlayerAgent * agent,
                                const double & lower_angle,
                                const double & upper_angle,
                                const bool clear_mode )
{
    double scan_range = upper_angle - lower_angle;
    double scan_dec = 8.0;
    double scan_count = rint( scan_range / scan_dec );
    scan_dec = scan_range / scan_count;

    rcsc::AngleDeg best_angle = 0.0;
    double best_score = 0.0;
    do
    {
        rcsc::AngleDeg tmp_angle = upper_angle - scan_range;

        double tmp_score = calc_score( agent, tmp_angle );
        if ( clear_mode )
        {
            tmp_score *= ( 0.5
                           * ( rcsc::AngleDeg::sin_deg(1.5 * tmp_angle.abs() + 45.0 )
                               + 1.0) );
        }
        int dir_count = agent->world().dirCount( tmp_angle ) - 3;
        if ( dir_count < 0 ) dir_count = 0;
        tmp_score *= std::pow( 0.95, static_cast< double >( dir_count ) );

        if ( tmp_score > best_score )
        {
            best_angle = tmp_angle;
            best_score = tmp_score;
        }
        scan_range -= scan_dec;

        rcsc::dlog.addText( rcsc::Logger::CLEAR,
                      "Body_ClearClone.get_best_angle. search_angle=%f, score=%f",
                      tmp_angle.degree(), tmp_score );

    }
    while ( scan_range >= 0.0 );

    return best_angle;
}

/*-------------------------------------------------------------------*/
/*!

*/
double
Body_ClearClone::calc_score( const rcsc::PlayerAgent * agent,
                            const rcsc::AngleDeg & target_angle )
{
    double score = 1.0;

    const rcsc::Line2D angle_line( agent->world().self().pos(),
                             target_angle );

    const rcsc::AngleDeg target_left_angle = target_angle - 30.0;
    const rcsc::AngleDeg target_right_angle = target_angle + 30.0;

    const rcsc::PlayerPtrCont::const_iterator end
        = agent->world().opponentsFromSelf().end();
    for ( rcsc::PlayerPtrCont::const_iterator
              it = agent->world().opponentsFromSelf().begin();
          it != end;
          ++it )
    {
        if ( (*it)->angleFromSelf().isWithin( target_left_angle,
                                              target_right_angle ) )
        {
            rcsc::Vector2D project_point = angle_line.projection( (*it)->pos() );
            double width = (*it)->pos().dist( project_point );
            double dist = agent->world().self().pos().dist( project_point );
            score *= width / dist;
        }
    }

    return score;
}

