#ifndef KUBO_INTERCEPTION_H
#define KUBO_INTERCEPTION_H

#include <rcsc/common/player_type.h>
#include <rcsc/geom/vector_2d.h>

//namespace rcsc {

/*!
  \class KuboInterception
  \brief interception estimation utility using Newton Method
*/
class KuboInterception {
public:
    static const double NEVER_CYCLE;

private:
    //! first ball position
    const rcsc::Vector2D M_ball_first_pos;
    //! first ball speed
    const double M_ball_first_speed;
    //! frist ball velocity direction
    const rcsc::AngleDeg M_ball_vel_angle;

    //! constant value for the Newton Method
    const double M_ball_x_constant;
    //! constant value for the Newton Method
    const double M_ball_x_d_constant;

public:

    /*
      \brief create with ball info
      \param ball_pos ball first poitiion
      \param ball_first_speed ball first speed
      \param ball_vel_angle ball velocity direction

      When we use this utility in pass route evaluation,
      next cycle info should be set to ball_pos and ball_vel.
    */
    KuboInterception( const rcsc::Vector2D & ball_pos,
                  const double & ball_first_speed,
                  const rcsc::AngleDeg & ball_vel_angle );

    /*
      \brief create with ball info
      \param ball_pos ball first poitiion
      \param ball_vel ball first velocity

      When we use this utility in pass route evaluation,
      next cycle info should be set to ball_pos and ball_vel.
    */
    KuboInterception( const rcsc::Vector2D & ball_pos,
                  const rcsc::Vector2D & ball_vel );

    /*!
      \brief predict ball gettable cycle for player using Newton Method
      \param player_pos player's position
      \param player_vel pointer to player velocity. if unknown, set NULL
      \param player_angle pointer to player angle. if unknown, set NULL
      \param player_count player's confidence count
      \param control_buf player's control area buffer (kickable area or catchable area)
      \param player_max_speed player's reachable max speed. player type paramter
      \return gettable cycle

      When we use this utility in pass or shoot route evaluation,
      next cycle info is set to player_pos.
    */
//     double getReachCycle( const rcsc::Vector2D & player_pos,
//                           const rcsc::Vector2D * player_vel,
//                           const rcsc::AngleDeg * player_angle,
//                           const int player_count,
//                           const double & control_buf,
//                           const double & player_max_speed ) const;

    /*!
      \brief predict ball gettable cycle for player using greedy method
      \param player_pos player's position
      \param player_vel pointer to player velocity. if unknown, set NULL
      \param player_angle pointer to player angle. if unknown, set NULL
      \param player_count player's confidence count
      \param control_buf player's control area buffer (kickable area or catchable area)
      \param player_max_speed player's reachable max speed. player type paramter

      When we use this utility in pass or shoot route evaluation,
      next cycle info is set to player_pos.
     */
//     double getReachCycleGreedly( const rcsc::Vector2D & player_pos,
//                                  const rcsc::Vector2D * player_vel, // if unknown, NULL
//                                  const rcsc::AngleDeg * player_angle, // if unknown, NULL
//                                  const int player_count,
//                                  const double & control_buf,
//                                  const double & player_max_speed ) const; // if unknown, NULL
    
    //////////////////added by okubo//////////////////////
    double getReachCycleKubo( const rcsc::Vector2D & player_pos,
                              const rcsc::Vector2D * player_vel,
                              const rcsc::AngleDeg * player_angle,
                              const double player_decay,
                              const double inertia_moment,
                              const double dash_max_power,
                              const int player_count,
                              const double & control_buf,
                              const double & player_max_speed) const;
    ///////////////////////////////////////////////////////

};

//}

#endif
