// -*-c++-*-

/*!
  \file kick_table.h
  \brief kick info holder Header File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef RCSC_PLAYER_KICK_TABLE_H
#define RCSC_PLAYER_KICK_TABLE_H

#include <rcsc/geom/vector_2d.h>

namespace rcsc {

class WorldModel;

/*-------------------------------------------------------------------*/
/*
  \class KickPoint
  \brief kick subtarget point info class
*/
class KickPoint {
public:
    static const int MAX_INTERFERE; //!< maximal interfere cycle

private:
    std::size_t M_index; //!< index of this point
    Vector2D M_ball_pos; //!< ball global coordinates
    double M_kick_rate; //!< kick rate
    /*!
      estimated opponent reach or pitch out cycle.
      it this value equals 0, this state is invalid.
     */
    int M_interfere_cycle;

    //! not used
    KickPoint();
public:
    //! constructor
    KickPoint( const std::size_t index,
               const Vector2D & ball_pos,
               const double & kick_rate,
               const int cycle )
        : M_index( index )
        , M_ball_pos( ball_pos )
        , M_kick_rate( kick_rate )
        , M_interfere_cycle( cycle )
      { }

    //! accesor
    std::size_t getIndex() const
      {
          return M_index;
      }

    //! accessor
    const
    Vector2D & ballPos() const
      {
          return M_ball_pos;
      }
    //! accessor
    const
    double & kickRate() const
      {
          return M_kick_rate;
      }
    //! accessor
    int interfereCycle() const
      {
          return M_interfere_cycle;
      }
};

typedef std::vector< KickPoint > KickPointCont;
typedef std::vector< const KickPoint * > KickPointPtrCont;

class TwoKick {
private:
    std::size_t M_first;
    std::size_t M_second;
    double M_score;
public:
    TwoKick( const std::size_t first,
             const std::size_t first,
             const double & score )
        : M_first( first )
        , M_sencod( second )
        , M_score( score )
      { }

};

/*-------------------------------------------------------------------*/
/*!
  \class KickTable
  \brief This class stores and updates latest kick subtargets.

  When player updates his world model, he also update his kick status.
  First, all possible subtargets are created.
  Second, all subtargets are sorted by opponet players' minimal reach cycle.
  This value is used as the criteria of kick action chain.
*/
class KickTable {
public:
    enum {
        MAX_DEPTH = 2,
        ANGLE_DIVS_LOW = 8,
        ANGLE_DIVS_HIGH = 12
    };
private:

    /*!
      the vector of subtarget kick points.
      this variable used and refered only from this class.
    */
    KickPointCont M_kick_points[MAX_DEPTH];
    /*!
      const pointer to the vector of subtarget kick points.
      this variable is refered from external modules.
    */
    KickPointPtrCont M_kick_points_ptr[MAX_DEPTH];

    // noncopyable
    Kicktable( const KickTable & );
    KickTable & operator=( const KickTable & );
public:
    //! constructor
    KickTable();

    //! generate kick point vector
    void update( const WorldModel & world );

    /*!
      accessor to the data container.
      NOTE: no throw for excepton.
      check the range before this method called.
      \param depth kick state depth. this must be within [0, MAX_DEPT-1]
    */
    inline
    const
    KickPointPtrCont * getKickPoints( const int depth ) const
      {
          if ( depth < 0 || MAX_DEPTH <= depth )
          {
              return static_cast< KickPointPtrCont* >( 0 );
          }
          return &( M_kick_points_ptr[depth] );
      }

private:
    //! set point and kick rate
    void setPoints( const WorldModel & world );
    //! check points validation
    int calcInterfereCycle( const WorldModel & world,
                            const Vector2D & point,
                            const int depth );

    //! sort kick points by interfere cycle
    void sortPoints();

};

}

#endif
